/*----- PROTECTED REGION ID(CbpmMaster.cpp) ENABLED START -----*/
static const char *RcsId = "$Id: CbpmMaster.cpp,v 1.8 2018-04-24 15:32:51 giulio Exp $";
//=============================================================================
//
// file :        CbpmMaster.cpp
//
// description : C++ source for the CbpmMaster and its commands.
//               The class is derived from Device. It represents the
//               CORBA servant object which will be accessed from the
//               network. All commands which can be executed on the
//               CbpmMaster are implemented in this file.
//
// project :     Cbpm Master.
//
// $Author: giulio $
//
// $Revision: 1.8 $
// $Date: 2018-04-24 15:32:51 $
//
// SVN only:
// $HeadURL:  $
//
// CVS only:
// $Source: /home/cvsadm/cvsroot/fermi/servers/cbpmmaster/src/CbpmMaster.cpp,v $
// $Log: CbpmMaster.cpp,v $
// Revision 1.8  2018-04-24 15:32:51  giulio
// Added Agc/Avc decimation values
//
// Revision 1.7  2018-04-19 08:17:56  giulio
// Automatic reset of the lantronix switches through the pdu's
//
// Revision 1.6  2014-05-07 12:10:39  giulio
// Persistent force command int lantronix switch logic. Bits high when FEL1
//
// Revision 1.5  2014-05-05 13:18:12  giulio
// Fixed some indexes
//
// Revision 1.4  2014-05-05 07:33:57  giulio
// Added rtbpm switch
//
// Revision 1.3  2013-08-22 14:20:40  giulio
// Moved to Tango8
//
// Revision 1.2  2012-04-20 12:16:33  giulio
// Fel1/fel2 switch based on init
//
// Revision 1.1.1.1  2012-04-16 12:45:50  giulio
// First release
//
//
//=============================================================================
//                This file is generated by POGO
//        (Program Obviously used to Generate tango Object)
//=============================================================================


#include <CbpmMaster.h>
#include <CbpmMasterClass.h>
#include <vector>
#include "cbpmmthread.h"

/*----- PROTECTED REGION END -----*/	//	CbpmMaster.cpp

/**
 *  CbpmMaster class description:
 *    Enable/disable cbpm's depending on fel01/fel02 configuration
 */

//================================================================
//  The following table gives the correspondence
//  between command and method names.
//
//  Command name       |  Method name
//================================================================
//  State              |  Inherited (no method)
//  Status             |  Inherited (no method)
//  ForceFel01         |  force_fel01
//  ForceFel02         |  force_fel02
//  GetCbpmFel01Names  |  get_cbpm_fel01_names
//  GetCbpmFel02Names  |  get_cbpm_fel02_names
//  GetBpmFel01Names   |  get_bpm_fel01_names
//  GetBpmFel02Names   |  get_bpm_fel02_names
//================================================================

//================================================================
//  Attributes managed are:
//================================================================
//  Enabled          |  Tango::DevBoolean	Scalar
//  Fel01            |  Tango::DevBoolean	Scalar
//  Fel02            |  Tango::DevBoolean	Scalar
//  AvcValMin        |  Tango::DevDouble	Scalar
//  AvcValMax        |  Tango::DevDouble	Scalar
//  AgcValMin        |  Tango::DevDouble	Scalar
//  AgcValMax        |  Tango::DevDouble	Scalar
//  Gain             |  Tango::DevDouble	Scalar
//  CbpmFel01States  |  Tango::DevState	Spectrum  ( max = 50)
//  CbpmFel02States  |  Tango::DevState	Spectrum  ( max = 50)
//  BpmFel01States   |  Tango::DevState	Spectrum  ( max = 50)
//  BpmFel02States   |  Tango::DevState	Spectrum  ( max = 50)
//================================================================

namespace CbpmMaster_ns
{
/*----- PROTECTED REGION ID(CbpmMaster::namespace_starting) ENABLED START -----*/

	//	static initializations

	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::namespace_starting

//--------------------------------------------------------
/**
 *	Method      : CbpmMaster::CbpmMaster()
 *	Description : Constructors for a Tango device
 *                implementing the classCbpmMaster
 */
//--------------------------------------------------------
CbpmMaster::CbpmMaster(Tango::DeviceClass *cl, string &s)
 : TANGO_BASE_CLASS(cl, s.c_str())
{
	/*----- PROTECTED REGION ID(CbpmMaster::constructor_1) ENABLED START -----*/

	init_device();

	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::constructor_1
}
//--------------------------------------------------------
CbpmMaster::CbpmMaster(Tango::DeviceClass *cl, const char *s)
 : TANGO_BASE_CLASS(cl, s)
{
	/*----- PROTECTED REGION ID(CbpmMaster::constructor_2) ENABLED START -----*/

	init_device();

	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::constructor_2
}
//--------------------------------------------------------
CbpmMaster::CbpmMaster(Tango::DeviceClass *cl, const char *s, const char *d)
 : TANGO_BASE_CLASS(cl, s, d)
{
	/*----- PROTECTED REGION ID(CbpmMaster::constructor_3) ENABLED START -----*/

	init_device();

	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::constructor_3
}

//--------------------------------------------------------
/**
 *	Method      : CbpmMaster::delete_device()
 *	Description : will be called at device destruction or at init command
 */
//--------------------------------------------------------
void CbpmMaster::delete_device()
{
	DEBUG_STREAM << "CbpmMaster::delete_device() " << device_name << endl;
	/*----- PROTECTED REGION ID(CbpmMaster::delete_device) ENABLED START -----*/

	//	Delete device allocated objects
	run_flag = false;
	while (run_thread == true) {
		sleep(1);
	}
	disconnect_devices();

	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::delete_device
	delete[] attr_Enabled_read;
	delete[] attr_Fel01_read;
	delete[] attr_Fel02_read;
	delete[] attr_AvcValMin_read;
	delete[] attr_AvcValMax_read;
	delete[] attr_AgcValMin_read;
	delete[] attr_AgcValMax_read;
	delete[] attr_Gain_read;
	delete[] attr_CbpmFel01States_read;
	delete[] attr_CbpmFel02States_read;
	delete[] attr_BpmFel01States_read;
	delete[] attr_BpmFel02States_read;
}

//--------------------------------------------------------
/**
 *	Method      : CbpmMaster::init_device()
 *	Description : will be called at device initialization.
 */
//--------------------------------------------------------
void CbpmMaster::init_device()
{
	DEBUG_STREAM << "CbpmMaster::init_device() create device " << device_name << endl;
	/*----- PROTECTED REGION ID(CbpmMaster::init_device_before) ENABLED START -----*/

	//	Initialization before get_device_property() call

	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::init_device_before
	

	//	Get the device properties from database
	get_device_property();
	if (mandatoryNotDefined)
		return;
	
	attr_Enabled_read = new Tango::DevBoolean[1];
	attr_Fel01_read = new Tango::DevBoolean[1];
	attr_Fel02_read = new Tango::DevBoolean[1];
	attr_AvcValMin_read = new Tango::DevDouble[1];
	attr_AvcValMax_read = new Tango::DevDouble[1];
	attr_AgcValMin_read = new Tango::DevDouble[1];
	attr_AgcValMax_read = new Tango::DevDouble[1];
	attr_Gain_read = new Tango::DevDouble[1];
	attr_CbpmFel01States_read = new Tango::DevState[50];
	attr_CbpmFel02States_read = new Tango::DevState[50];
	attr_BpmFel01States_read = new Tango::DevState[50];
	attr_BpmFel02States_read = new Tango::DevState[50];

	/*----- PROTECTED REGION ID(CbpmMaster::init_device) ENABLED START -----*/

	//	Initialize device
	num_cbpm_fel01 = 0;
	num_cbpm_fel02 = 0;
	num_bpm_fel01 = 0;
	num_bpm_fel02 = 0;	

	bpm_fel01_switch_error = 0;
	bpm_fel02_switch_error = 0;
	switch_fel01_counter = 0;
	switch_fel02_counter = 0;

	*attr_AvcValMin_read = 1000;
	*attr_AvcValMax_read = 30000;

	*attr_AgcValMin_read = 1000;
	*attr_AgcValMax_read = 30000;

	*attr_Gain_read = 35;

	gettimeofday(&last_thread_time, NULL);
	last_thread_time.tv_sec -= 60;

	connect_devices();

	run_thread = false;	
	run_flag = true;	
	force_fel01_flag = false;
	force_fel02_flag = false;
	cbpmmloop = new cbpmmthread::cbpmmthread(this);
	INFO_STREAM << "Start monitor loop" << endl;
	cbpmmloop->start();	
	INFO_STREAM << "Initialization done" << endl;

	attr_Enabled_write = false;

	switch_running = false;
	switch_fel01_counter = 0;
	switch_fel02_counter = 0;

	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::init_device
}

//--------------------------------------------------------
/**
 *	Method      : CbpmMaster::get_device_property()
 *	Description : Read database to initialize property data members.
 */
//--------------------------------------------------------
void CbpmMaster::get_device_property()
{
	/*----- PROTECTED REGION ID(CbpmMaster::get_device_property_before) ENABLED START -----*/

	//	Initialize property data members

	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::get_device_property_before

	mandatoryNotDefined = false;

	//	Read device properties from database.
	Tango::DbData	dev_prop;
	dev_prop.push_back(Tango::DbDatum("PlcDevice"));
	dev_prop.push_back(Tango::DbDatum("Fel01CbpmDevices"));
	dev_prop.push_back(Tango::DbDatum("Fel02CbpmDevices"));
	dev_prop.push_back(Tango::DbDatum("CheckPeriod"));
	dev_prop.push_back(Tango::DbDatum("SwitchTcpPort"));
	dev_prop.push_back(Tango::DbDatum("SwitchTcpAddress"));
	dev_prop.push_back(Tango::DbDatum("Fel01BpmDevices"));
	dev_prop.push_back(Tango::DbDatum("Fel02BpmDevices"));
	dev_prop.push_back(Tango::DbDatum("SwitchPDU"));

	//	is there at least one property to be read ?
	if (dev_prop.size()>0)
	{
		//	Call database and extract values
		if (Tango::Util::instance()->_UseDb==true)
			get_db_device()->get_property(dev_prop);
	
		//	get instance on CbpmMasterClass to get class property
		Tango::DbDatum	def_prop, cl_prop;
		CbpmMasterClass	*ds_class =
			(static_cast<CbpmMasterClass *>(get_device_class()));
		int	i = -1;

		//	Try to initialize PlcDevice from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  plcDevice;
		else {
			//	Try to initialize PlcDevice from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  plcDevice;
		}
		//	And try to extract PlcDevice value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  plcDevice;

		//	Try to initialize Fel01CbpmDevices from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  fel01CbpmDevices;
		else {
			//	Try to initialize Fel01CbpmDevices from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  fel01CbpmDevices;
		}
		//	And try to extract Fel01CbpmDevices value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  fel01CbpmDevices;

		//	Try to initialize Fel02CbpmDevices from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  fel02CbpmDevices;
		else {
			//	Try to initialize Fel02CbpmDevices from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  fel02CbpmDevices;
		}
		//	And try to extract Fel02CbpmDevices value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  fel02CbpmDevices;

		//	Try to initialize CheckPeriod from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  checkPeriod;
		else {
			//	Try to initialize CheckPeriod from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  checkPeriod;
		}
		//	And try to extract CheckPeriod value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  checkPeriod;

		//	Try to initialize SwitchTcpPort from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  switchTcpPort;
		else {
			//	Try to initialize SwitchTcpPort from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  switchTcpPort;
		}
		//	And try to extract SwitchTcpPort value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  switchTcpPort;
		//	Property StartDsPath is mandatory, check if has been defined in database.
		check_mandatory_property(cl_prop, dev_prop[i]);

		//	Try to initialize SwitchTcpAddress from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  switchTcpAddress;
		else {
			//	Try to initialize SwitchTcpAddress from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  switchTcpAddress;
		}
		//	And try to extract SwitchTcpAddress value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  switchTcpAddress;
		//	Property StartDsPath is mandatory, check if has been defined in database.
		check_mandatory_property(cl_prop, dev_prop[i]);

		//	Try to initialize Fel01BpmDevices from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  fel01BpmDevices;
		else {
			//	Try to initialize Fel01BpmDevices from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  fel01BpmDevices;
		}
		//	And try to extract Fel01BpmDevices value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  fel01BpmDevices;

		//	Try to initialize Fel02BpmDevices from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  fel02BpmDevices;
		else {
			//	Try to initialize Fel02BpmDevices from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  fel02BpmDevices;
		}
		//	And try to extract Fel02BpmDevices value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  fel02BpmDevices;

		//	Try to initialize SwitchPDU from class property
		cl_prop = ds_class->get_class_property(dev_prop[++i].name);
		if (cl_prop.is_empty()==false)	cl_prop  >>  switchPDU;
		else {
			//	Try to initialize SwitchPDU from default device value
			def_prop = ds_class->get_default_device_property(dev_prop[i].name);
			if (def_prop.is_empty()==false)	def_prop  >>  switchPDU;
		}
		//	And try to extract SwitchPDU value from database
		if (dev_prop[i].is_empty()==false)	dev_prop[i]  >>  switchPDU;

	}

	/*----- PROTECTED REGION ID(CbpmMaster::get_device_property_after) ENABLED START -----*/

	//	Check device property data members init

	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::get_device_property_after
}
//--------------------------------------------------------
/**
 *	Method      : CbpmMaster::check_mandatory_property()
 *	Description : For mandatory properties check if defined in database.
 */
//--------------------------------------------------------
void CbpmMaster::check_mandatory_property(Tango::DbDatum &class_prop, Tango::DbDatum &dev_prop)
{
	//	Check if all properties are empty
	if (class_prop.is_empty() && dev_prop.is_empty())
	{
		TangoSys_OMemStream	tms;
		tms << endl <<"Property \'" << dev_prop.name;
		if (Tango::Util::instance()->_UseDb==true)
			tms << "\' is mandatory but not defined in database";
		else
			tms << "\' is mandatory but cannot be defined without database";
		string	status(get_status());
		status += tms.str();
		set_status(status);
		mandatoryNotDefined = true;
		/*----- PROTECTED REGION ID(CbpmMaster::check_mandatory_property) ENABLED START -----*/
		cerr << tms.str() << " for " << device_name << endl;
		
		/*----- PROTECTED REGION END -----*/	//	CbpmMaster::check_mandatory_property
	}
}


//--------------------------------------------------------
/**
 *	Method      : CbpmMaster::always_executed_hook()
 *	Description : method always executed before any command is executed
 */
//--------------------------------------------------------
void CbpmMaster::always_executed_hook()
{
	INFO_STREAM << "CbpmMaster::always_executed_hook()  " << device_name << endl;
	if (mandatoryNotDefined)
	{
		string	status(get_status());
		Tango::Except::throw_exception(
					(const char *)"PROPERTY_NOT_SET",
					status.c_str(),
					(const char *)"CbpmMaster::always_executed_hook()");
	}
	/*----- PROTECTED REGION ID(CbpmMaster::always_executed_hook) ENABLED START -----*/

	//	code always executed before all requests
	set_DevStat();

	for (unsigned int i = 0; i < switchTcpAddress.size(); i++) {
		struct timeval curr_state_time;
		gettimeofday(&curr_state_time, NULL);

		if (((curr_state_time.tv_sec - lantronix[i].socket_state_time.tv_sec) > checkPeriod) && (switch_running == false)) {
			if (lantronix[i].sock)
				close(lantronix[i].sock);
			sock_connect(i);
			gettimeofday(&lantronix[i].socket_state_time,NULL);
		}

	}

	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::always_executed_hook
}

//--------------------------------------------------------
/**
 *	Method      : CbpmMaster::read_attr_hardware()
 *	Description : Hardware acquisition for attributes
 */
//--------------------------------------------------------
void CbpmMaster::read_attr_hardware(TANGO_UNUSED(vector<long> &attr_list))
{
	DEBUG_STREAM << "CbpmMaster::read_attr_hardware(vector<long> &attr_list) entering... " << endl;
	/*----- PROTECTED REGION ID(CbpmMaster::read_attr_hardware) ENABLED START -----*/

	//	Add your own code

	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::read_attr_hardware
}
//--------------------------------------------------------
/**
 *	Method      : CbpmMaster::write_attr_hardware()
 *	Description : Hardware writing for attributes
 */
//--------------------------------------------------------
void CbpmMaster::write_attr_hardware(TANGO_UNUSED(vector<long> &attr_list))
{
	DEBUG_STREAM << "CbpmMaster::write_attr_hardware(vector<long> &attr_list) entering... " << endl;
	/*----- PROTECTED REGION ID(CbpmMaster::write_attr_hardware) ENABLED START -----*/
	
	//	Add your own code
	
	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::write_attr_hardware
}

//--------------------------------------------------------
/**
 *	Read attribute Enabled related method
 *	Description: Enable automatic switching
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CbpmMaster::read_Enabled(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CbpmMaster::read_Enabled(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(CbpmMaster::read_Enabled) ENABLED START -----*/

	//	Set the attribute value
	*attr_Enabled_read = attr_Enabled_write;
	attr.set_value(attr_Enabled_read);

	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::read_Enabled
}
//--------------------------------------------------------
/**
 *	Write attribute Enabled related method
 *	Description: Enable automatic switching
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CbpmMaster::write_Enabled(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "CbpmMaster::write_Enabled(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevBoolean	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(CbpmMaster::write_Enabled) ENABLED START -----*/
	attr_Enabled_write = w_val;
	

	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::write_Enabled
}
//--------------------------------------------------------
/**
 *	Read attribute Fel01 related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CbpmMaster::read_Fel01(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CbpmMaster::read_Fel01(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(CbpmMaster::read_Fel01) ENABLED START -----*/

	//	Set the attribute value
	attr.set_value(attr_Fel01_read);

	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::read_Fel01
}
//--------------------------------------------------------
/**
 *	Read attribute Fel02 related method
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CbpmMaster::read_Fel02(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CbpmMaster::read_Fel02(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(CbpmMaster::read_Fel02) ENABLED START -----*/

	//	Set the attribute value
	attr.set_value(attr_Fel02_read);

	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::read_Fel02
}
//--------------------------------------------------------
/**
 *	Read attribute AvcValMin related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CbpmMaster::read_AvcValMin(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CbpmMaster::read_AvcValMin(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(CbpmMaster::read_AvcValMin) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_AvcValMin_read);
	
	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::read_AvcValMin
}
//--------------------------------------------------------
/**
 *	Write attribute AvcValMin related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CbpmMaster::write_AvcValMin(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "CbpmMaster::write_AvcValMin(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(CbpmMaster::write_AvcValMin) ENABLED START -----*/
	*attr_AvcValMin_read = w_val;
	
	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::write_AvcValMin
}
//--------------------------------------------------------
/**
 *	Read attribute AvcValMax related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CbpmMaster::read_AvcValMax(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CbpmMaster::read_AvcValMax(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(CbpmMaster::read_AvcValMax) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_AvcValMax_read);
	
	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::read_AvcValMax
}
//--------------------------------------------------------
/**
 *	Write attribute AvcValMax related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CbpmMaster::write_AvcValMax(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "CbpmMaster::write_AvcValMax(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(CbpmMaster::write_AvcValMax) ENABLED START -----*/
	*attr_AvcValMax_read = w_val;
	
	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::write_AvcValMax
}
//--------------------------------------------------------
/**
 *	Read attribute AgcValMin related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CbpmMaster::read_AgcValMin(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CbpmMaster::read_AgcValMin(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(CbpmMaster::read_AgcValMin) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_AgcValMin_read);
	
	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::read_AgcValMin
}
//--------------------------------------------------------
/**
 *	Write attribute AgcValMin related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CbpmMaster::write_AgcValMin(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "CbpmMaster::write_AgcValMin(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(CbpmMaster::write_AgcValMin) ENABLED START -----*/
	*attr_AgcValMin_read = w_val;
	
	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::write_AgcValMin
}
//--------------------------------------------------------
/**
 *	Read attribute AgcValMax related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CbpmMaster::read_AgcValMax(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CbpmMaster::read_AgcValMax(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(CbpmMaster::read_AgcValMax) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_AgcValMax_read);
	
	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::read_AgcValMax
}
//--------------------------------------------------------
/**
 *	Write attribute AgcValMax related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CbpmMaster::write_AgcValMax(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "CbpmMaster::write_AgcValMax(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(CbpmMaster::write_AgcValMax) ENABLED START -----*/
	*attr_AgcValMax_read = w_val;
	
	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::write_AgcValMax
}
//--------------------------------------------------------
/**
 *	Read attribute Gain related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CbpmMaster::read_Gain(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CbpmMaster::read_Gain(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(CbpmMaster::read_Gain) ENABLED START -----*/
	//	Set the attribute value
	attr.set_value(attr_Gain_read);
	
	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::read_Gain
}
//--------------------------------------------------------
/**
 *	Write attribute Gain related method
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
//--------------------------------------------------------
void CbpmMaster::write_Gain(Tango::WAttribute &attr)
{
	DEBUG_STREAM << "CbpmMaster::write_Gain(Tango::WAttribute &attr) entering... " << endl;
	//	Retrieve write value
	Tango::DevDouble	w_val;
	attr.get_write_value(w_val);
	/*----- PROTECTED REGION ID(CbpmMaster::write_Gain) ENABLED START -----*/
	*attr_Gain_read = w_val;
	
	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::write_Gain
}
//--------------------------------------------------------
/**
 *	Read attribute CbpmFel01States related method
 *	Description: 
 *
 *	Data type:	Tango::DevState
 *	Attr type:	Spectrum max = 50
 */
//--------------------------------------------------------
void CbpmMaster::read_CbpmFel01States(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CbpmMaster::read_CbpmFel01States(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(CbpmMaster::read_CbpmFel01States) ENABLED START -----*/

	//	Set the attribute value
	
	for (int i = 0; i < num_cbpm_fel01; i++) {
		attr_CbpmFel01States_read[i] = cbpm_fel01[i].state;
	}
	attr.set_value(attr_CbpmFel01States_read, num_cbpm_fel01);

	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::read_CbpmFel01States
}
//--------------------------------------------------------
/**
 *	Read attribute CbpmFel02States related method
 *	Description: 
 *
 *	Data type:	Tango::DevState
 *	Attr type:	Spectrum max = 50
 */
//--------------------------------------------------------
void CbpmMaster::read_CbpmFel02States(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CbpmMaster::read_CbpmFel02States(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(CbpmMaster::read_CbpmFel02States) ENABLED START -----*/

	//	Set the attribute value
	
	for (int i = 0; i < num_cbpm_fel02; i++) {
		attr_CbpmFel02States_read[i] = cbpm_fel02[i].state;
	}
	attr.set_value(attr_CbpmFel02States_read, num_cbpm_fel02);

	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::read_CbpmFel02States
}
//--------------------------------------------------------
/**
 *	Read attribute BpmFel01States related method
 *	Description: 
 *
 *	Data type:	Tango::DevState
 *	Attr type:	Spectrum max = 50
 */
//--------------------------------------------------------
void CbpmMaster::read_BpmFel01States(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CbpmMaster::read_BpmFel01States(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(CbpmMaster::read_BpmFel01States) ENABLED START -----*/
	//	Set the attribute value
	
	for (int i = 0; i < num_bpm_fel01; i++) {
		attr_BpmFel01States_read[i] = bpm_fel01[i].state;
	}
	attr.set_value(attr_BpmFel01States_read, num_bpm_fel01);
	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::read_BpmFel01States
}
//--------------------------------------------------------
/**
 *	Read attribute BpmFel02States related method
 *	Description: 
 *
 *	Data type:	Tango::DevState
 *	Attr type:	Spectrum max = 50
 */
//--------------------------------------------------------
void CbpmMaster::read_BpmFel02States(Tango::Attribute &attr)
{
	DEBUG_STREAM << "CbpmMaster::read_BpmFel02States(Tango::Attribute &attr) entering... " << endl;
	/*----- PROTECTED REGION ID(CbpmMaster::read_BpmFel02States) ENABLED START -----*/
	//	Set the attribute value
	
	for (int i = 0; i < num_bpm_fel02; i++) {
		attr_BpmFel02States_read[i] = bpm_fel02[i].state;
	}
	attr.set_value(attr_BpmFel02States_read, num_bpm_fel02);
	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::read_BpmFel02States
}

//--------------------------------------------------------
/**
 *	Method      : CbpmMaster::add_dynamic_attributes()
 *	Description : Create the dynamic attributes if any
 *                for specified device.
 */
//--------------------------------------------------------
void CbpmMaster::add_dynamic_attributes()
{
	/*----- PROTECTED REGION ID(CbpmMaster::add_dynamic_attributes) ENABLED START -----*/
	
	//	Add your own code to create and add dynamic attributes if any
	
	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::add_dynamic_attributes
}

//--------------------------------------------------------
/**
 *	Command ForceFel01 related method
 *	Description: 
 *
 */
//--------------------------------------------------------
void CbpmMaster::force_fel01()
{
	DEBUG_STREAM << "CbpmMaster::ForceFel01()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(CbpmMaster::force_fel01) ENABLED START -----*/

	//	Add your own code
	if (attr_Enabled_write) {
		Tango::Except::throw_exception (
			(const char *)"Automatic setting still enabled",
			(const char *)"Failed to force fel01 configuration",
			(const char *)"CbpmMaster::force_fel01()");		
	}
	
	if (!force_fel02_flag) {
		force_fel01_flag = true;
	}
	else {
		Tango::Except::throw_exception (
			(const char *)"Already forcing fel02",
			(const char *)"Failed to force fel01 configuration",
			(const char *)"CbpmMaster::force_fel01()");			
	}
	
	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::force_fel01
}
//--------------------------------------------------------
/**
 *	Command ForceFel02 related method
 *	Description: 
 *
 */
//--------------------------------------------------------
void CbpmMaster::force_fel02()
{
	DEBUG_STREAM << "CbpmMaster::ForceFel02()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(CbpmMaster::force_fel02) ENABLED START -----*/

	//	Add your own code
	if (attr_Enabled_write) {
		Tango::Except::throw_exception (
			(const char *)"Automatic setting still enabled",
			(const char *)"Failed to force fel02 configuration",
			(const char *)"CbpmMaster::force_fel02()");		
	}
	if (!force_fel01_flag) {
		force_fel02_flag = true;
	}
	else {
		Tango::Except::throw_exception (
			(const char *)"Already forcing fel02",
			(const char *)"Failed to force fel01 configuration",
			(const char *)"CbpmMaster::force_fel02()");	
	}
	
	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::force_fel02
}
//--------------------------------------------------------
/**
 *	Command GetCbpmFel01Names related method
 *	Description: 
 *
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarStringArray *CbpmMaster::get_cbpm_fel01_names()
{
	Tango::DevVarStringArray *argout;
	DEBUG_STREAM << "CbpmMaster::GetCbpmFel01Names()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(CbpmMaster::get_cbpm_fel01_names) ENABLED START -----*/

	//	Add your own code
	argout  = new Tango::DevVarStringArray();
	argout->length(num_cbpm_fel01);
	
	for (unsigned int i = 0; i < argout->length(); i++) {
		(*argout)[i] = CORBA::string_dup(cbpm_fel01[i].label.c_str());
	}

	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::get_cbpm_fel01_names
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetCbpmFel02Names related method
 *	Description: 
 *
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarStringArray *CbpmMaster::get_cbpm_fel02_names()
{
	Tango::DevVarStringArray *argout;
	DEBUG_STREAM << "CbpmMaster::GetCbpmFel02Names()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(CbpmMaster::get_cbpm_fel02_names) ENABLED START -----*/

	//	Add your own code
	argout  = new Tango::DevVarStringArray();
	argout->length(num_cbpm_fel02);
	
	for (unsigned int i = 0; i < argout->length(); i++) {
		(*argout)[i] = CORBA::string_dup(cbpm_fel02[i].label.c_str());
	}

	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::get_cbpm_fel02_names
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetBpmFel01Names related method
 *	Description: 
 *
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarStringArray *CbpmMaster::get_bpm_fel01_names()
{
	Tango::DevVarStringArray *argout;
	DEBUG_STREAM << "CbpmMaster::GetBpmFel01Names()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(CbpmMaster::get_bpm_fel01_names) ENABLED START -----*/
	//	Add your own code
	argout  = new Tango::DevVarStringArray();
	argout->length(num_bpm_fel01);
	
	for (unsigned int i = 0; i < argout->length(); i++) {
		(*argout)[i] = CORBA::string_dup(bpm_fel01[i].label.c_str());
	}
	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::get_bpm_fel01_names
	return argout;
}
//--------------------------------------------------------
/**
 *	Command GetBpmFel02Names related method
 *	Description: 
 *
 *	@returns 
 */
//--------------------------------------------------------
Tango::DevVarStringArray *CbpmMaster::get_bpm_fel02_names()
{
	Tango::DevVarStringArray *argout;
	DEBUG_STREAM << "CbpmMaster::GetBpmFel02Names()  - " << device_name << endl;
	/*----- PROTECTED REGION ID(CbpmMaster::get_bpm_fel02_names) ENABLED START -----*/
	//	Add your own code
	argout  = new Tango::DevVarStringArray();
	argout->length(num_bpm_fel02);
	
	for (unsigned int i = 0; i < argout->length(); i++) {
		(*argout)[i] = CORBA::string_dup(bpm_fel02[i].label.c_str());
	}
	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::get_bpm_fel02_names
	return argout;
}

/*----- PROTECTED REGION ID(CbpmMaster::namespace_ending) ENABLED START -----*/

	//	Additional Methods

//+------------------------------------------------------------------
/**
 *	method:	CbpmMaster::connect_devices
 * Connect to devices
*/
//+------------------------------------------------------------------
void CbpmMaster::connect_devices(void)
{
	
	num_cbpm_fel01 = fel01CbpmDevices.size();
	num_cbpm_fel02 = fel02CbpmDevices.size();

	num_pdu = switchPDU.size();

	for (int i = 0; i < num_cbpm_fel01; i++) {
		/* connect to the cbpm */
		cbpm_fel01[i].cbpmdevice = 0; 
		cbpm_fel01[i].devname = fel01CbpmDevices[i];
		if ((cbpm_fel01[i].cbpmdevice = new Tango::DeviceProxy(cbpm_fel01[i].devname)) == 0) {
			ERROR_STREAM << "Failed to get device " << cbpm_fel01[i].devname << " from database" << endl;
		}
		try {
			cbpm_fel01[i].cbpmdevice->set_timeout_millis(3000);
		}
		catch(Tango::DevFailed &e) {
			ERROR_STREAM << "Error setting timeout on " << cbpm_fel01[i].devname << endl;
		}
		vector<string> results; 
		string_explode(cbpm_fel01[i].devname, "/", &results);
		if (results.size() == 3) {
			cbpm_fel01[i].label = results[2];
		}
		else {
			ERROR_STREAM << "Invalid device name " << cbpm_fel01[i].devname << endl;
		}
		cbpm_fel01[i].state = (Tango::UNKNOWN);
		gettimeofday(&cbpm_fel01[i].state_time,NULL);	
	}	
				
	for (int i = 0; i < num_cbpm_fel02; i++) {
		/* connect to the cbpm */
		cbpm_fel02[i].cbpmdevice = 0;
		cbpm_fel02[i].devname = fel02CbpmDevices[i];
		if ((cbpm_fel02[i].cbpmdevice = new Tango::DeviceProxy(cbpm_fel02[i].devname)) == 0) {
			ERROR_STREAM << "Failed to get device " << cbpm_fel02[i].devname << " from database" << endl;	
		}
		try {
			cbpm_fel02[i].cbpmdevice->set_timeout_millis(3000);
		}
		catch(Tango::DevFailed &e) {
			ERROR_STREAM << "Error setting timeout on " << cbpm_fel02[i].devname << endl;
		}
		vector<string> results; 
		string_explode(cbpm_fel02[i].devname, "/", &results);
		if (results.size() == 3) {
			cbpm_fel02[i].label = results[2];
		}
		else {
			ERROR_STREAM << "Invalid device name " << cbpm_fel02[i].devname << endl;
		}
		cbpm_fel02[i].state = (Tango::UNKNOWN);
		gettimeofday(&cbpm_fel02[i].state_time,NULL);	
	}	

	plcdevice = 0;
	if ((plcdevice = new Tango::DeviceProxy(plcDevice)) == 0) {
		ERROR_STREAM << "Failed to get plc device " << plcDevice << endl;	
	}
	plc_state = (Tango::UNKNOWN);

	num_bpm_fel01 = fel01BpmDevices.size();
	num_bpm_fel02 = fel02BpmDevices.size();

	for (int i = 0; i < num_bpm_fel01; i++) {
		/* connect to the cbpm */
		bpm_fel01[i].bpmdevice = 0; 
		bpm_fel01[i].devname = fel01BpmDevices[i];
		if ((bpm_fel01[i].bpmdevice = new Tango::DeviceProxy(bpm_fel01[i].devname)) == 0) {
			ERROR_STREAM << "Failed to get device " << bpm_fel01[i].devname << " from database" << endl;
		}
		try {
			bpm_fel01[i].bpmdevice->set_timeout_millis(3000);
		}
		catch(Tango::DevFailed &e) {
			ERROR_STREAM << "Error setting timeout on " << bpm_fel01[i].devname << endl;
		}
		vector<string> results; 
		string_explode(bpm_fel01[i].devname, "/", &results);
		if (results.size() == 3) {
			bpm_fel01[i].label = results[2];
		}
		else {
			ERROR_STREAM << "Invalid device name " << bpm_fel01[i].devname << endl;
		}
		bpm_fel01[i].state = (Tango::UNKNOWN);
		gettimeofday(&bpm_fel01[i].state_time,NULL);	
	}	
				
	for (int i = 0; i < num_bpm_fel02; i++) {
		/* connect to the cbpm */
		bpm_fel02[i].bpmdevice = 0;
		bpm_fel02[i].devname = fel02BpmDevices[i];
		if ((bpm_fel02[i].bpmdevice = new Tango::DeviceProxy(bpm_fel02[i].devname)) == 0) {
			ERROR_STREAM << "Failed to get device " << bpm_fel02[i].devname << " from database" << endl;	
		}
		try {
			bpm_fel02[i].bpmdevice->set_timeout_millis(3000);
		}
		catch(Tango::DevFailed &e) {
			ERROR_STREAM << "Error setting timeout on " << bpm_fel02[i].devname << endl;
		}
		vector<string> results; 
		string_explode(bpm_fel02[i].devname, "/", &results);
		if (results.size() == 3) {
			bpm_fel02[i].label = results[2];
		}
		else {
			ERROR_STREAM << "Invalid device name " << bpm_fel02[i].devname << endl;
		}
		bpm_fel02[i].state = (Tango::UNKNOWN);
		gettimeofday(&bpm_fel02[i].state_time,NULL);	
	}

	for (int i = 0; i < num_pdu; i++) {
		/* connect to the cbpm */
		pdu[i].pdudevice = 0;
		vector<string> results; 
		string_explode(switchPDU[i], ";", &results);
		if (results.size() == 2) {
			pdu[i].devname = results[0];
			pdu[i].outlet = atoi(results[1].c_str());
		}
		else {
			ERROR_STREAM << "PDU configuration error: " << switchPDU[i] << endl;
			num_pdu = 0;
		}
		if ((pdu[i].pdudevice = new Tango::DeviceProxy(pdu[i].devname)) == 0) {
			ERROR_STREAM << "Failed to get device " << pdu[i].devname << " from database" << endl;	
			num_pdu = 0;
		}
		try {
			pdu[i].pdudevice->set_timeout_millis(3000);
		}
		catch(Tango::DevFailed &e) {
			ERROR_STREAM << "Error setting timeout on " << pdu[i].devname << endl;
			num_pdu = 0;
		}
		pdu[i].state = (Tango::UNKNOWN);
		gettimeofday(&pdu[i].state_time,NULL);	
	}

	/* set tcp socket time */
	for (unsigned int i = 0; i < switchTcpAddress.size(); i++) {
		gettimeofday(&lantronix[i].socket_state_time,NULL);
		lantronix[i].mutex = new omni_mutex::omni_mutex();	
		lantronix[i].sock = 0;
		lantronix[i].state = Tango::UNKNOWN;
	}



}

//+------------------------------------------------------------------
/**
 *	method:	CbpmMaster::disconnect_devices
 *  Disconnect from devices
*/
//+------------------------------------------------------------------
void CbpmMaster::disconnect_devices(void)
{

	for (int i = 0; i < num_cbpm_fel01; i++) {
		/* delete cbpm devices */
		if (cbpm_fel01[i].cbpmdevice) {
			delete cbpm_fel01[i].cbpmdevice;
		}
	}
	for (int i = 0; i < num_cbpm_fel02; i++) {
		/* delete cbpm devices */
		if (cbpm_fel02[i].cbpmdevice) {
			delete cbpm_fel02[i].cbpmdevice;
		}
	}

	for (int i = 0; i < num_bpm_fel01; i++) {
		/* delete cbpm devices */
		if (bpm_fel01[i].bpmdevice) {
			delete bpm_fel01[i].bpmdevice;
		}
	}
	for (int i = 0; i < num_bpm_fel02; i++) {
		/* delete cbpm devices */
		if (bpm_fel02[i].bpmdevice) {
			delete bpm_fel02[i].bpmdevice;
		}
	}

	for (int i = 0; i < num_pdu; i++) {
		/* delete cbpm devices */
		if (pdu[i].pdudevice) {
			delete pdu[i].pdudevice;
		}
	}

	for (unsigned int i = 0; i < switchTcpAddress.size(); i++) {
		delete lantronix[i].mutex;
		if (lantronix[i].sock)
			close(lantronix[i].sock);
	}

	if (plcdevice)
		delete plcdevice;

}


//+------------------------------------------------------------------
/**
 *	method:	CbpmMaster::string_explode
 * Explode a string 
*/
//+------------------------------------------------------------------
void CbpmMaster::string_explode(string str, string separator, vector<string>* results)
{
	unsigned int found;
	
	found = str.find_first_of(separator);
	while(found != string::npos) {
		if(found > 0) {
			results->push_back(str.substr(0,found));
		}
		str = str.substr(found+1);
		found = str.find_first_of(separator);
	}
	if(str.length() > 0) {
		results->push_back(str);
	}

}


//+------------------------------------------------------------------
/**
 *	method:	CbpmMaster::set_DevStat
 * Explode a string 
*/
//+------------------------------------------------------------------
void CbpmMaster::set_DevStat(void)
{
	struct timeval time_now;
	long diff_time;
	std::stringstream s;
	int device_error;	

	gettimeofday(&time_now, NULL);
	
	diff_time = abs((time_now.tv_sec - last_thread_time.tv_sec)*1000000 +
		(time_now.tv_usec - last_thread_time.tv_usec));
	
	set_state(Tango::ON);

	if (switch_running) {
		set_state(Tango::MOVING);
		s << "MOVING: Switching bpms ";
		if (switch_fel01) {
			int num_bpm = num_cbpm_fel01+num_bpm_fel01;
			s << "(" << switch_fel01_counter << "/" << num_bpm << ")" << endl;
		}
		else {
			int num_bpm = num_cbpm_fel02+num_bpm_fel02;
			s << "(" << switch_fel02_counter << "/" << num_bpm << ")" << endl;
		}
	}
	
	if ((*attr_Fel01_read == false) && (*attr_Fel02_read == false)) {
		set_state(Tango::ALARM);
		s << "ALARM: FEL unknown state" << endl;			
	}
	else {
		if(*attr_Fel01_read) {
			s << "FEL01 enabled" << endl;	
		}
		else if (*attr_Fel02_read) {
			s << "FEL02 enabled" << endl;
		}
	}


	/* check if plc comm is ok */
	if (plc_state == (Tango::FAULT)) {
		set_state(Tango::FAULT);
		s << "ERROR: pcl communication error" << endl;	
	}

	/* check if thread is running or not */
	if ((diff_time/1000000) > 5*(num_cbpm_fel01+num_cbpm_fel02+num_bpm_fel01+num_bpm_fel02)) {
		set_state(Tango::FAULT);
		s << "ERROR: checkthread not running" << endl;
	}

	/* check pdu state */
	device_error = 0;
	for (int i = 0; i < num_pdu; i++) {
		if (pdu[i].state == (Tango::OFF)) {
			s << "Resetting pdu " << pdu[i].devname << ", outlet " << pdu[i].outlet << endl;
		}
		else if (pdu[i].state == (Tango::FAULT)) {
			s << "ERROR setting pdu " << pdu[i].devname << ", outlet " << pdu[i].outlet << endl;
		}
	}
	if (device_error) {
		set_state(Tango::FAULT);
	}

	/* check lantronix state */
	device_error = 0;
	for (int i = 0; i < switchTcpAddress.size(); i++) {
		if (lantronix[i].state != Tango::ON) {
			s << "ERROR: " << switchTcpAddress[i] << " communication error (lantronix switch)" << endl;
			device_error++;
		}
	}
	if (device_error) {
		set_state(Tango::FAULT);
	}

	/* check bpm states */
	device_error = 0;
	for (int i = 0; i < num_cbpm_fel01; i++) {
		if ((cbpm_fel01[i].state == Tango::FAULT) || (cbpm_fel01[i].state == Tango::UNKNOWN)) {
			s << "ERROR: " << cbpm_fel01[i].devname << " fault (state=" << cbpm_fel01[i].state << ")" << endl;
			device_error++;
		}
		if ((cbpm_fel01[i].state == Tango::ALARM)) {
			s << "ALARM: " << cbpm_fel01[i].devname << endl;
		}
	}
	for (int i = 0; i < num_cbpm_fel02; i++) {
		if ((cbpm_fel02[i].state == Tango::FAULT) || (cbpm_fel02[i].state == Tango::UNKNOWN)) {
			s << "ERROR: " << cbpm_fel02[i].devname << " fault (state=" << cbpm_fel02[i].state << ")" << endl;
			device_error++;
		}
		if ((cbpm_fel02[i].state == Tango::ALARM)) {
			s << "ALARM: " << cbpm_fel02[i].devname << endl;
		}
	}
	for (int i = 0; i < num_bpm_fel01; i++) {
		if ((bpm_fel01[i].state == Tango::FAULT) || (bpm_fel01[i].state == Tango::UNKNOWN)) {
			s << "ERROR: " << bpm_fel01[i].devname << " fault (state=" << bpm_fel01[i].state << ")" << endl;
			device_error++;
		}
		if ((bpm_fel01[i].state == Tango::ALARM)) {
			s << "ALARM: " << bpm_fel01[i].devname << endl;
		}
	}
	for (int i = 0; i < num_bpm_fel02; i++) {
		if ((bpm_fel02[i].state == Tango::FAULT) || (bpm_fel02[i].state == Tango::UNKNOWN)) {
			s << "ERROR: " << bpm_fel02[i].devname << " fault (state=" << bpm_fel02[i].state << ")" << endl;
			device_error++;
		}
		if ((bpm_fel02[i].state == Tango::ALARM)) {
			s << "ALARM: " << bpm_fel02[i].devname << endl;
		}
	}
	if (device_error) {
		set_state(Tango::FAULT);
	}

	/* check if all bpms switched to fel01 */
	device_error = 0;
	if (bpm_fel01_switch_error > 4) {
		if (*attr_Fel01_read) {
			for (int i = 0; i < num_cbpm_fel01; i++) {
				if ((cbpm_fel01[i].state != Tango::ON) && (cbpm_fel01[i].state != Tango::OFF) &&
					(cbpm_fel01[i].state != Tango::ALARM)) {
					s << "ERROR: " << cbpm_fel01[i].devname << " failed to switch to FEL1 (state=" << cbpm_fel01[i].state << ")" << endl;
					device_error++;
				}
			}
			for (int i = 0; i < num_bpm_fel01; i++) {
				if ((bpm_fel01[i].state != Tango::ON) && (bpm_fel01[i].state != Tango::OFF) &&
					(cbpm_fel01[i].state != Tango::ALARM)) {
					s << "ERROR: " << bpm_fel01[i].devname << " failed to switch to FEL1 (state=" << bpm_fel01[i].state << ")" << endl;
					device_error++;
				}
			}
		}
	}
	if (device_error) {
		set_state(Tango::FAULT);
	}

	/* check if all bpms switched to fel02 */
	device_error = 0;
	if (bpm_fel02_switch_error > 4) {
		if (*attr_Fel02_read) {
			for (int i = 0; i < num_cbpm_fel02; i++) {
				if ((cbpm_fel02[i].state != Tango::ON) && (cbpm_fel02[i].state != Tango::OFF) &&
					(cbpm_fel02[i].state != Tango::ALARM)) {
					s << "ERROR: " << cbpm_fel02[i].devname << " failed to switch to FEL2 (state=" << cbpm_fel02[i].state << ")"<< endl;
					device_error++;
				}
			}
			for (int i = 0; i < num_bpm_fel02; i++) {
				if ((bpm_fel02[i].state != Tango::ON) && (bpm_fel02[i].state != Tango::OFF) && 
					(bpm_fel02[i].state != Tango::ALARM)) {
					s << "ERROR: " << bpm_fel02[i].devname << " failed to switch to FEL2 (state=" << bpm_fel02[i].state << ")" << endl;
					device_error++;
				}
			}
		}
	}
	if (device_error) {
		set_state(Tango::FAULT);
	}
	
	set_status(s.str());

}


//+------------------------------------------------------------------
/**
 *      method: CbpmMaster::::sock_connect()
 *
 */
//+------------------------------------------------------------------
int CbpmMaster::sock_connect(int idx)
{
	int connect_err_flag = 0,arg;
	TangoSys_MemStream out_stream;

	lantronix[idx].mutex->lock();

	if (lantronix[idx].sock)
			close(lantronix[idx].sock);

	INFO_STREAM << "Connecting device " << switchTcpAddress[idx] << endl;

	lantronix[idx].mysocket.sin_port = htons( switchTcpPort[idx]);
	lantronix[idx].mysocket.sin_addr.s_addr = inet_addr(switchTcpAddress[idx].c_str());
	lantronix[idx].mysocket.sin_family = AF_INET;
	if ((lantronix[idx].sock = socket(AF_INET, SOCK_STREAM, 0)) < 0 ) {
		connect_err_flag++;
		INFO_STREAM << "Open device " << switchTcpAddress[idx] << " failed, " << strerror(errno) << endl;
		lantronix[idx].state = Tango::FAULT;
		set_status(out_stream.str());
		lantronix[idx].mutex->unlock();
		return connect_err_flag;
	}

	 /* getting socket parameters */
	if( (arg = fcntl(lantronix[idx].sock, F_GETFL, NULL)) < 0) {
		connect_err_flag++;
		INFO_STREAM << "Reading socket setting " << switchTcpAddress[idx] << " failed, " << strerror(errno) << endl;
		lantronix[idx].state = Tango::FAULT;
		set_status(out_stream.str());
		lantronix[idx].mutex->unlock();
		return connect_err_flag;
	}

	 /* setting non blocking socket */
	arg |= O_NONBLOCK;
	if (fcntl (lantronix[idx].sock, F_SETFL, arg) < 0) {
		connect_err_flag++;
		INFO_STREAM << "Setting socket " << switchTcpAddress[idx] << " failed, " << strerror(errno) << endl;
		lantronix[idx].state = Tango::FAULT;
		set_status(out_stream.str());
		lantronix[idx].mutex->unlock();
		return connect_err_flag;
	}

	/* connect to device */
	/* do not generate exception because non blocking socket configured */
	connect (lantronix[idx].sock, (const sockaddr*)&lantronix[idx].mysocket, sizeof(struct sockaddr));
	lantronix[idx].state = Tango::ON;


	INFO_STREAM << "Connected to " << switchTcpAddress[idx] << endl;
	sock_flush(idx);

	lantronix[idx].mutex->unlock();

	return connect_err_flag;

}

//+------------------------------------------------------------------
/**
 *      method: CbpmMaster::::sock_flush
 *
 */
//+------------------------------------------------------------------
void CbpmMaster::sock_flush(int idx)
{
	char trash[LANTRONIX_MSG_LEN];
	int count=0;

	while (recv(lantronix[idx].sock, trash, LANTRONIX_MSG_LEN, 0) >= 0) {
		count++;
		if (count > 10)
			break;
	}      /* flush socket */

}

//+------------------------------------------------------------------
/**
 *      method: CbpmMaster::::write_cmd
 *
 */
//+------------------------------------------------------------------
void CbpmMaster::write_cmd(int idx, int fel)
{

	if ((fel == 1) || (fel == 2)) {

		char strin[LANTRONIX_MSG_LEN], strout[LANTRONIX_MSG_LEN];
		int num;

		lantronix[idx].mutex->lock();
		sock_flush(idx);

		strin[0]=LANTRONIX_CMD_WRITE;
		strin[1]=MASKALLBIT;
		strin[2]=0x0; strin[3]=0x0;strin[4]=0x0;
		if(fel == 1) 
			strin[5]=MASKALLBIT;
		else 
			strin[5]=0x0;
		strin[6]=0x0; strin[7]=0x0; strin[8]=0x0;

		//printf("%d %d %d %d %d %d %d %d %d\n",
		//	strin[0],strin[1],strin[2],strin[3],strin[4],strin[5],strin[6],strin[7],strin[8]);

		if ((num = write(lantronix[idx].sock, strin, 9)) < 0) {
			TangoSys_MemStream out_stream; out_stream << strerror(errno) << ends;
			INFO_STREAM << "socket write error (" << switchTcpAddress[idx] << "): " << out_stream.str() << endl;
			lantronix[idx].state = (Tango::FAULT);
			lantronix[idx].mutex->unlock();
			return;
		}

		omni_thread::sleep(0,100000000); // convert millisecond to nanosecond

		memset(strout,0,LANTRONIX_MSG_LEN);
		if ((num = recv(lantronix[idx].sock, strout, 5, 0)) < 0) {
			TangoSys_MemStream out_stream; out_stream << strerror(errno) << ends;
			INFO_STREAM << "socket read error (" << switchTcpAddress[idx] << "): " << out_stream.str() << endl;
			lantronix[idx].state = (Tango::FAULT); 
			lantronix[idx].mutex->unlock();
			return;
		}

		//printf("%d %d %d %d %d\n",
		//	strout[0],strout[1],strout[2],strout[3],strout[4]);

		if (((fel == 2) && (strout[1] != 0)) || ((fel == 1) && (strout[1] != MASKALLBIT))) {
			INFO_STREAM << "switch error (" << switchTcpAddress[idx] << "): " << "switch command failed" << endl;
			lantronix[idx].state = (Tango::FAULT); 
			lantronix[idx].mutex->unlock();
			return;
		}

		lantronix[idx].state = (Tango::ON);
		sock_flush(idx);
		lantronix[idx].mutex->unlock();

	}

	gettimeofday(&lantronix[idx].socket_state_time,NULL);

}


//+------------------------------------------------------------------
/**
 *      method: CbpmMaster::::pdu_off
 *
 */
//+------------------------------------------------------------------
void CbpmMaster::pdu_off(int idx)
{
	Tango::DeviceData Din;

	try {
		Din << pdu[idx].outlet;
		pdu[idx].pdudevice->command_inout("OutletOff", Din);
		pdu[idx].state = (Tango::OFF);
		INFO_STREAM << "Setting pdu off " << pdu[idx].devname << ", outlet " << pdu[idx].outlet << endl;
	}
	catch(Tango::DevFailed &e) {
		ERROR_STREAM << "Error setting pdu off " << pdu[idx].devname << ", outlet " << pdu[idx].outlet << endl;
		pdu[idx].state = (Tango::FAULT);
	}

}


//+------------------------------------------------------------------
/**
 *      method: CbpmMaster::::pdu_on
 *
 */
//+------------------------------------------------------------------
void CbpmMaster::pdu_on(int idx)
{
	Tango::DeviceData Din;

	try {
		Din << pdu[idx].outlet;
		pdu[idx].pdudevice->command_inout("OutletOn", Din);
		pdu[idx].state = (Tango::ON);
		INFO_STREAM << "Setting pdu on " << pdu[idx].devname << ", outlet " << pdu[idx].outlet << endl;
	}
	catch(Tango::DevFailed &e) {
		ERROR_STREAM << "Error setting pdu on " << pdu[idx].devname << ", outlet " << pdu[idx].outlet << endl;
		pdu[idx].state = (Tango::FAULT);
	}

}



	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::namespace_ending
} //	namespace
