/*----- PROTECTED REGION ID(CbpmMaster.h) ENABLED START -----*/
//=============================================================================
//
// file :        CbpmMaster.h
//
// description : Include for the CbpmMaster class.
//
// project :     Cbpm Master.
//
// $Author: giulio $
//
// $Revision: 1.5 $
// $Date: 2018-04-24 15:32:51 $
//
// SVN only:
// $HeadURL:  $
//
// CVS only:
// $Source: /home/cvsadm/cvsroot/fermi/servers/cbpmmaster/src/CbpmMaster.h,v $
// $Log: CbpmMaster.h,v $
// Revision 1.5  2018-04-24 15:32:51  giulio
// Added Agc/Avc decimation values
//
// Revision 1.4  2018-04-19 08:17:56  giulio
// Automatic reset of the lantronix switches through the pdu's
//
// Revision 1.3  2014-05-05 13:18:12  giulio
// Fixed some indexes
//
// Revision 1.2  2014-05-05 07:33:57  giulio
// Added rtbpm switch
//
// Revision 1.1.1.1  2012-04-16 12:45:50  giulio
// First release
//
//
//=============================================================================
//                This file is generated by POGO
//        (Program Obviously used to Generate tango Object)
//=============================================================================


#ifndef CBPMMASTER_H
#define CBPMMASTER_H

#include <tango.h>

#include <sys/time.h>
#include <time.h>

#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define CBPMMASTER_TOUT 10
#define CBPMMASTER_DEV_TOUT_MILLIS 1000

#define CBPMMASTER_NUM_MAX_BPM  50

#define LANTRONIX_DEV_TOUT        1000 /* ms. */
#define LANTRONIX_MSG_LEN      100
#define LANTRONIX_RECONNECT_MAXERR 20 /* sec */

#define LANTRONIX_MAX_NUM  10 

#define LANTRONIX_CMD_WRITE   0x1B  

#define MASKBIT1 0x01
#define MASKBIT2 0x02
#define MASKBIT3 0x04

#define MASKALLBIT 0x07


/*----- PROTECTED REGION END -----*/	//	CbpmMaster.h

/**
 *  CbpmMaster class description:
 *    Enable/disable cbpm's depending on fel01/fel02 configuration
 */

namespace CbpmMaster_ns
{
/*----- PROTECTED REGION ID(CbpmMaster::Additional Class Declarations) ENABLED START -----*/

		//		Additional Class Declarations
	class cbpmmthread;
	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::Additional Class Declarations

class CbpmMaster : public TANGO_BASE_CLASS
{

/*----- PROTECTED REGION ID(CbpmMaster::Data Members) ENABLED START -----*/

	//		Add your own data members
	friend class cbpmmthread;

	typedef struct {
		string devname,
		       label;
		Tango::DeviceProxy *cbpmdevice;
		Tango::DevState state;
		struct timeval state_time;	
	} cbpm_t;	

	typedef struct {
		string devname,
		       label;
		Tango::DeviceProxy *bpmdevice;
		Tango::DevState state;
		struct timeval state_time;	
	} bpm_t;

	typedef struct {
		struct sockaddr_in mysocket;
		int sock;
		omni_mutex::omni_mutex *mutex;
		struct timeval socket_state_time;
		Tango::DevState state;
	} socket_t;

	typedef struct {
		string devname,
		       label;
		Tango::DeviceProxy *pdudevice;
		Tango::DevState state;
		struct timeval state_time;	
		Tango::DevUShort outlet;
	} pdu_t;

	long num_cbpm_fel01, num_cbpm_fel02, num_bpm_fel01, num_bpm_fel02, num_pdu;
	long 	bpm_fel01_switch_error, bpm_fel02_switch_error;


	cbpm_t cbpm_fel01[CBPMMASTER_NUM_MAX_BPM], cbpm_fel02[CBPMMASTER_NUM_MAX_BPM];
	bpm_t bpm_fel01[CBPMMASTER_NUM_MAX_BPM], bpm_fel02[CBPMMASTER_NUM_MAX_BPM];

	socket_t lantronix[LANTRONIX_MAX_NUM];
	pdu_t pdu[LANTRONIX_MAX_NUM];

	Tango::DeviceProxy *plcdevice;
	Tango::DevState plc_state;

	cbpmmthread *cbpmmloop;
	struct timeval last_thread_time;
	bool run_flag, run_thread;
	bool force_fel01_flag, force_fel02_flag;
		
	Tango::DevBoolean attr_Enabled_write;

	bool switch_running, switch_fel01;
	long switch_fel01_counter, switch_fel02_counter;

	struct sockaddr_in proto_socket;

	/*
	struct sockaddr_in mysocket[LANTRONIX_MAX_NUM];
	int sock[LANTRONIX_MAX_NUM];
	omni_mutex::omni_mutex *mutex[LANTRONIX_MAX_NUM];
	struct timeval socket_state_time[LANTRONIX_MAX_NUM];
	*/

	void write_cmd(int idx, int fel);
	int sock_connect(int i);
	void sock_flush(int i);
	void pdu_off(int idx);
	void pdu_on(int idx);

	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::Data Members

//	Device property data members
public:
	//	PlcDevice:	Plc device name used to get information about fel01/02 status
	string	plcDevice;
	//	Fel01CbpmDevices:	Fel01 cbpm devices
	vector<string>	fel01CbpmDevices;
	//	Fel02CbpmDevices:	Fel02 cbpm devices
	vector<string>	fel02CbpmDevices;
	//	CheckPeriod:	seconds
	Tango::DevLong	checkPeriod;
	//	SwitchTcpPort:	
	vector<Tango::DevLong>	switchTcpPort;
	//	SwitchTcpAddress:	
	vector<string>	switchTcpAddress;
	//	Fel01BpmDevices:	Fel01 bpm devices
	vector<string>	fel01BpmDevices;
	//	Fel02BpmDevices:	Fel02 bpm devices
	vector<string>	fel02BpmDevices;
	//	SwitchPDU:	
	vector<string>	switchPDU;

	bool	mandatoryNotDefined;

//	Attribute data members
public:
	Tango::DevBoolean	*attr_Enabled_read;
	Tango::DevBoolean	*attr_Fel01_read;
	Tango::DevBoolean	*attr_Fel02_read;
	Tango::DevDouble	*attr_AvcValMin_read;
	Tango::DevDouble	*attr_AvcValMax_read;
	Tango::DevDouble	*attr_AgcValMin_read;
	Tango::DevDouble	*attr_AgcValMax_read;
	Tango::DevDouble	*attr_Gain_read;
	Tango::DevState	*attr_CbpmFel01States_read;
	Tango::DevState	*attr_CbpmFel02States_read;
	Tango::DevState	*attr_BpmFel01States_read;
	Tango::DevState	*attr_BpmFel02States_read;

//	Constructors and destructors
public:
	/**
	 * Constructs a newly device object.
	 *
	 *	@param cl	Class.
	 *	@param s 	Device Name
	 */
	CbpmMaster(Tango::DeviceClass *cl,string &s);
	/**
	 * Constructs a newly device object.
	 *
	 *	@param cl	Class.
	 *	@param s 	Device Name
	 */
	CbpmMaster(Tango::DeviceClass *cl,const char *s);
	/**
	 * Constructs a newly device object.
	 *
	 *	@param cl	Class.
	 *	@param s 	Device name
	 *	@param d	Device description.
	 */
	CbpmMaster(Tango::DeviceClass *cl,const char *s,const char *d);
	/**
	 * The device object destructor.
	 */	
	~CbpmMaster() {delete_device();};


//	Miscellaneous methods
public:
	/*
	 *	will be called at device destruction or at init command.
	 */
	void delete_device();
	/*
	 *	Initialize the device
	 */
	virtual void init_device();
	/*
	 *	Read the device properties from database
	 */
	void get_device_property();
	/*
	 *	Always executed method before execution command method.
	 */
	virtual void always_executed_hook();

	/*
	 *	Check if mandatory property has been set
	 */
	 void check_mandatory_property(Tango::DbDatum &class_prop, Tango::DbDatum &dev_prop);

//	Attribute methods
public:
	//--------------------------------------------------------
	/*
	 *	Method      : CbpmMaster::read_attr_hardware()
	 *	Description : Hardware acquisition for attributes.
	 */
	//--------------------------------------------------------
	virtual void read_attr_hardware(vector<long> &attr_list);
	//--------------------------------------------------------
	/*
	 *	Method      : CbpmMaster::write_attr_hardware()
	 *	Description : Hardware writing for attributes.
	 */
	//--------------------------------------------------------
	virtual void write_attr_hardware(vector<long> &attr_list);

/**
 *	Attribute Enabled related methods
 *	Description: Enable automatic switching
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
	virtual void read_Enabled(Tango::Attribute &attr);
	virtual void write_Enabled(Tango::WAttribute &attr);
	virtual bool is_Enabled_allowed(Tango::AttReqType type);
/**
 *	Attribute Fel01 related methods
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
	virtual void read_Fel01(Tango::Attribute &attr);
	virtual bool is_Fel01_allowed(Tango::AttReqType type);
/**
 *	Attribute Fel02 related methods
 *	Description: 
 *
 *	Data type:	Tango::DevBoolean
 *	Attr type:	Scalar
 */
	virtual void read_Fel02(Tango::Attribute &attr);
	virtual bool is_Fel02_allowed(Tango::AttReqType type);
/**
 *	Attribute AvcValMin related methods
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_AvcValMin(Tango::Attribute &attr);
	virtual void write_AvcValMin(Tango::WAttribute &attr);
	virtual bool is_AvcValMin_allowed(Tango::AttReqType type);
/**
 *	Attribute AvcValMax related methods
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_AvcValMax(Tango::Attribute &attr);
	virtual void write_AvcValMax(Tango::WAttribute &attr);
	virtual bool is_AvcValMax_allowed(Tango::AttReqType type);
/**
 *	Attribute AgcValMin related methods
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_AgcValMin(Tango::Attribute &attr);
	virtual void write_AgcValMin(Tango::WAttribute &attr);
	virtual bool is_AgcValMin_allowed(Tango::AttReqType type);
/**
 *	Attribute AgcValMax related methods
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_AgcValMax(Tango::Attribute &attr);
	virtual void write_AgcValMax(Tango::WAttribute &attr);
	virtual bool is_AgcValMax_allowed(Tango::AttReqType type);
/**
 *	Attribute Gain related methods
 *	Description: 
 *
 *	Data type:	Tango::DevDouble
 *	Attr type:	Scalar
 */
	virtual void read_Gain(Tango::Attribute &attr);
	virtual void write_Gain(Tango::WAttribute &attr);
	virtual bool is_Gain_allowed(Tango::AttReqType type);
/**
 *	Attribute CbpmFel01States related methods
 *	Description: 
 *
 *	Data type:	Tango::DevState
 *	Attr type:	Spectrum max = 50
 */
	virtual void read_CbpmFel01States(Tango::Attribute &attr);
	virtual bool is_CbpmFel01States_allowed(Tango::AttReqType type);
/**
 *	Attribute CbpmFel02States related methods
 *	Description: 
 *
 *	Data type:	Tango::DevState
 *	Attr type:	Spectrum max = 50
 */
	virtual void read_CbpmFel02States(Tango::Attribute &attr);
	virtual bool is_CbpmFel02States_allowed(Tango::AttReqType type);
/**
 *	Attribute BpmFel01States related methods
 *	Description: 
 *
 *	Data type:	Tango::DevState
 *	Attr type:	Spectrum max = 50
 */
	virtual void read_BpmFel01States(Tango::Attribute &attr);
	virtual bool is_BpmFel01States_allowed(Tango::AttReqType type);
/**
 *	Attribute BpmFel02States related methods
 *	Description: 
 *
 *	Data type:	Tango::DevState
 *	Attr type:	Spectrum max = 50
 */
	virtual void read_BpmFel02States(Tango::Attribute &attr);
	virtual bool is_BpmFel02States_allowed(Tango::AttReqType type);


	//--------------------------------------------------------
	/**
	 *	Method      : CbpmMaster::add_dynamic_attributes()
	 *	Description : Add dynamic attributes if any.
	 */
	//--------------------------------------------------------
	void add_dynamic_attributes();



//	Command related methods
public:
	/**
	 *	Command ForceFel01 related method
	 *	Description: 
	 *
	 */
	virtual void force_fel01();
	virtual bool is_ForceFel01_allowed(const CORBA::Any &any);
	/**
	 *	Command ForceFel02 related method
	 *	Description: 
	 *
	 */
	virtual void force_fel02();
	virtual bool is_ForceFel02_allowed(const CORBA::Any &any);
	/**
	 *	Command GetCbpmFel01Names related method
	 *	Description: 
	 *
	 *	@returns 
	 */
	virtual Tango::DevVarStringArray *get_cbpm_fel01_names();
	virtual bool is_GetCbpmFel01Names_allowed(const CORBA::Any &any);
	/**
	 *	Command GetCbpmFel02Names related method
	 *	Description: 
	 *
	 *	@returns 
	 */
	virtual Tango::DevVarStringArray *get_cbpm_fel02_names();
	virtual bool is_GetCbpmFel02Names_allowed(const CORBA::Any &any);
	/**
	 *	Command GetBpmFel01Names related method
	 *	Description: 
	 *
	 *	@returns 
	 */
	virtual Tango::DevVarStringArray *get_bpm_fel01_names();
	virtual bool is_GetBpmFel01Names_allowed(const CORBA::Any &any);
	/**
	 *	Command GetBpmFel02Names related method
	 *	Description: 
	 *
	 *	@returns 
	 */
	virtual Tango::DevVarStringArray *get_bpm_fel02_names();
	virtual bool is_GetBpmFel02Names_allowed(const CORBA::Any &any);


/*----- PROTECTED REGION ID(CbpmMaster::Additional Method prototypes) ENABLED START -----*/

	//	Additional Method prototypes
	void connect_devices(void);
	void disconnect_devices(void);
	void string_explode(string str, string separator, vector<string>* results);
	void set_DevStat(void);

	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::Additional Method prototypes
};

/*----- PROTECTED REGION ID(CbpmMaster::Additional Classes Definitions) ENABLED START -----*/

	//	Additional Classes definitions

	/*----- PROTECTED REGION END -----*/	//	CbpmMaster::Additional Classes Definitions

}	//	End of namespace

#endif   //	CbpmMaster_H
