//
// file :         cbpmmthread.cpp
//
// description :  gige camera acquisition thread
//
// project :      TANGO Device Server
//
//
//
// copyleft :   Sincrotrone Trieste S.C.p.A.
//              Strada Statale 14 - km 163,5 in AREA Science Park
//              34012 Basovizza, Trieste ITALY
//

#include <omnithread.h>
#include "CbpmMaster.h"
#include "cbpmmthread.h"

namespace CbpmMaster_ns
{
/* 
 *	cbpmmthread::run()
 *	Run  
 */		
void cbpmmthread::run(void *) 
{
	vector <bool> plc_stat;
	struct timeval loop_start, loop_end;
	int loop_time_len, loop_time_diff; /*  usec */
	bool local_force_fel01_flag, local_force_fel02_flag; 
	int fel01_disabled, fel02_disabled;	
	int fel01_enabled, fel02_enabled;
	bool cmd_fel01_enable, cmd_fel02_enable;
	int fel01_unknown, fel02_unknown;
	int fel_value = 0;
	int lantronix_fel = 1;

	myacq->run_thread = true;
	
	/* read cbpm fel01 */
	while (myacq->run_flag) {

		fel01_disabled = 0; fel02_disabled = 0;
		fel01_enabled = 0; fel02_enabled = 0;
		fel01_unknown = 0; fel02_unknown = 0;
		cmd_fel01_enable = false; cmd_fel02_enable = false;


		gettimeofday(&loop_start, NULL);
	
		/************ read cbpm fel01 ************/
		for (int i = 0; i < myacq->num_cbpm_fel01; i++) {
			try {
				myacq->cbpm_fel01[i].cbpmdevice->read_attribute("State") >> myacq->cbpm_fel01[i].state;
				if ((myacq->cbpm_fel01[i].state == Tango::DISABLE) ||  (myacq->cbpm_fel01[i].state == Tango::FAULT)) {
					fel01_disabled++;
				}
				if ((myacq->cbpm_fel01[i].state != Tango::DISABLE) && (myacq->cbpm_fel01[i].state != Tango::FAULT)) {
					fel01_enabled++;
				}
			}
			catch(Tango::DevFailed e) {
				myacq->cbpm_fel01[i].state = Tango::UNKNOWN;
				fel01_unknown++;
			}			
			if (myacq->run_flag == false)
				break;
		}
		if (fel01_disabled * fel01_enabled)
			myacq->bpm_fel01_switch_error++;
		/*****************************************/

		/************ read cbpm fel02 ************/
		for (int i = 0; i < myacq->num_cbpm_fel02; i++) {
			try {
				myacq->cbpm_fel02[i].cbpmdevice->read_attribute("State") >> myacq->cbpm_fel02[i].state;
				if ((myacq->cbpm_fel02[i].state == Tango::DISABLE) || (myacq->cbpm_fel02[i].state == Tango::FAULT)) {
					fel02_disabled++;
				}
				if ((myacq->cbpm_fel02[i].state != Tango::DISABLE) && (myacq->cbpm_fel02[i].state != Tango::FAULT)) {
					fel02_enabled++;
				}
			}
			catch(Tango::DevFailed e) {
				myacq->cbpm_fel02[i].state = Tango::UNKNOWN;
				fel02_unknown++;
			}
			if (myacq->run_flag == false)
				break;
		}
		if (fel02_disabled * fel02_enabled)
			myacq->bpm_fel01_switch_error++;
		/*****************************************/

		/************ read bpm fel01 ************/
		for (int i = 0; i < myacq->num_bpm_fel01; i++) {
			try {
				myacq->bpm_fel01[i].bpmdevice->read_attribute("State") >> myacq->bpm_fel01[i].state;
				if ((myacq->bpm_fel01[i].state == Tango::DISABLE) ||  (myacq->bpm_fel01[i].state == Tango::FAULT)) {
					fel01_disabled++;
				}
				if ((myacq->bpm_fel01[i].state != Tango::DISABLE) && (myacq->bpm_fel01[i].state != Tango::FAULT)) {
					fel01_enabled++;
				}
			}
			catch(Tango::DevFailed e) {
				myacq->bpm_fel01[i].state = Tango::UNKNOWN;
				fel01_unknown++;
			}			
			if (myacq->run_flag == false)
				break;
		}
		/* if some bpms are enabled and others not */
		if (fel01_disabled * fel01_enabled)
			myacq->bpm_fel01_switch_error++;
		else
			myacq->bpm_fel01_switch_error = 0;
		/*****************************************/

		/************ read bpm fel02 ************/
		for (int i = 0; i < myacq->num_bpm_fel02; i++) {
			try {
				myacq->bpm_fel02[i].bpmdevice->read_attribute("State") >> myacq->bpm_fel02[i].state;
				if ((myacq->bpm_fel02[i].state == Tango::DISABLE) ||  (myacq->bpm_fel02[i].state == Tango::FAULT)) {
					fel02_disabled++;
				}
				if ((myacq->bpm_fel01[i].state != Tango::DISABLE) && (myacq->bpm_fel01[i].state != Tango::FAULT)) {
					fel02_enabled++;
				}
			}
			catch(Tango::DevFailed e) {
				myacq->bpm_fel02[i].state = Tango::UNKNOWN;
				fel01_unknown++;
			}			
			if (myacq->run_flag == false)
				break;
		}
		/* if some bpms are enabled and others not */
		if (fel02_disabled * fel02_enabled)
			myacq->bpm_fel02_switch_error++;
		else
			myacq->bpm_fel02_switch_error = 0;
		/*****************************************/

		/* read plc */
		if ((myacq->attr_Enabled_write) && myacq->plcdevice) {
			try {
				myacq->plcdevice->read_attribute("Out_system") >> plc_stat;
				myacq->plc_state = Tango::ON;
				if (plc_stat.size() == 192) {
					*myacq->attr_Fel01_read = plc_stat[4];
					*myacq->attr_Fel02_read = plc_stat[5];
				}
				else {
					myacq->plc_state = Tango::FAULT;
					*myacq->attr_Fel01_read = false;
					*myacq->attr_Fel02_read = false;
				}
			}		
			catch(Tango::DevFailed e) {
				myacq->plc_state = Tango::FAULT;
				*myacq->attr_Fel01_read = false;
				*myacq->attr_Fel02_read = false;
			}			
		}
		else {
			*myacq->attr_Fel01_read = false;
			*myacq->attr_Fel02_read = false;
		}		

		/* when fel01 or fel02 enabled */
		if ((*myacq->attr_Fel01_read != false) || (*myacq->attr_Fel02_read != false)) {
			/* if fel01 enabled */
			if (*myacq->attr_Fel01_read) {
				/* if incoherencies force again fel01 */
				if (fel01_disabled || fel02_enabled)
					cmd_fel01_enable = true;
			}
			/* if fel02 enabled */
			if (*myacq->attr_Fel02_read) {
				/* if incoherencies force again fel02 */
				if (fel01_enabled || fel02_disabled)
					cmd_fel02_enable = true;
			}
		}


		/* reset switch state if auto enabled and fel state defined, otherwise keep last value (forced or not)*/
		if (myacq->attr_Enabled_write && (*myacq->attr_Fel01_read || *myacq->attr_Fel02_read))
			lantronix_fel = 0;

		/*  reset lantronix via PDU */
		if (myacq->force_fel01_flag || myacq->force_fel02_flag) {
			for (int i = 0; i < myacq->num_pdu; i++) {
				myacq->pdu_off(i);
			}
			omni_thread::sleep(3,0);
			for (int i = 0; i < myacq->num_pdu; i++) {
				myacq->pdu_on(i);
			}
			omni_thread::sleep(5,0);

			for (unsigned int i = 0; i < myacq->switchTcpAddress.size(); i++) { 
				omni_thread::sleep(1,0);
				myacq->sock_connect(i);
			}

		}

		/******** set lantronix switch ********/
		if ((*myacq->attr_Fel01_read) || myacq->force_fel01_flag)
			lantronix_fel = 1;
		else if ((*myacq->attr_Fel02_read) || myacq->force_fel02_flag)
			lantronix_fel = 2;
		for (unsigned int i = 0; i < myacq->switchTcpAddress.size(); i++) 
			myacq->write_cmd(i,lantronix_fel);

		myacq->switch_running = false; /* to avoid socket timeout during fel switch */
		/******** end lantronix switch ********/

		if (cmd_fel01_enable && fel01_unknown) 
			cmd_fel01_enable = false;
		
		if (cmd_fel02_enable && fel02_unknown) 
			cmd_fel02_enable = false;

		gettimeofday(&myacq->last_thread_time,NULL);
		

		/************* ENABLE FEL01 *************/
		local_force_fel01_flag = myacq->force_fel01_flag;

		if (myacq->force_fel01_flag) myacq->force_fel01_flag = 0;
		if (cmd_fel01_enable || local_force_fel01_flag) {
			myacq->switch_running = true;
			myacq->switch_fel01_counter = 0;
			myacq->switch_fel02_counter = 0;
			myacq->switch_fel01 = true;
			bool enable = false;
			/************* disable cbpm fel_02 *************/
			for (int i = 0; i < myacq->num_cbpm_fel02; i++) {
				try {
					/* recover from lock */
					if(myacq->cbpm_fel02[i].state == Tango::ALARM) {
						myacq->cbpm_fel02[i].cbpmdevice->command_inout("Init");
						omni_thread::sleep(1,500000000);
					}
					Tango::DeviceAttribute fel02enabled("Fel02Enabled", enable);
					myacq->cbpm_fel02[i].cbpmdevice->write_attribute(fel02enabled);
					omni_thread::sleep(0,150000000);
				}
				catch(Tango::DevFailed e) {
					myacq->cbpm_fel02[i].state = Tango::UNKNOWN;										
				}
				if (myacq->run_flag == false)
					break;							
			}
			omni_thread::sleep(2,0);
			for (int i = 0; i < myacq->num_cbpm_fel02; i++) {
				try {
					myacq->cbpm_fel02[i].cbpmdevice->read_attribute("State") >> myacq->cbpm_fel02[i].state;
				}
				catch(Tango::DevFailed e) {
					myacq->cbpm_fel02[i].state = Tango::UNKNOWN;										
				}		
				if (myacq->run_flag == false)
					break;				
			}	
			/*********** end disable cbpm fel_02 ***********/

			/************* enable cbpm fel_01 *************/
			for (int i = 0; i < myacq->num_cbpm_fel01; i++) {
				myacq->switch_fel01_counter++;
				try {
					myacq->cbpm_fel01[i].cbpmdevice->read_attribute("State") >> myacq->cbpm_fel01[i].state;
				}
				catch(Tango::DevFailed e) {}
				try {
					myacq->cbpm_fel01[i].cbpmdevice->command_inout("Init");
					omni_thread::sleep(1,500000000);
					Tango::DeviceAttribute fel02enabled("Fel02Enabled", false);
					myacq->cbpm_fel01[i].cbpmdevice->write_attribute(fel02enabled);
					omni_thread::sleep(1,500000000);
					myacq->cbpm_fel01[i].cbpmdevice->read_attribute("State") >> myacq->cbpm_fel01[i].state;
				}
				catch(Tango::DevFailed e) {
					myacq->cbpm_fel01[i].state = Tango::UNKNOWN;
				}	
				if (myacq->run_flag == false)
					break;
			}
			/*********** end enable cbpm fel_01 ***********/

			/************* disable bpm fel_02 *************/
			for (int i = 0; i < myacq->num_bpm_fel02; i++) {
				try {
					myacq->bpm_fel02[i].bpmdevice->command_inout("Disable");
					omni_thread::sleep(1,500000000);
					myacq->bpm_fel02[i].bpmdevice->read_attribute("State") >> myacq->bpm_fel02[i].state;
				}
				catch(Tango::DevFailed e) {
					myacq->bpm_fel02[i].state = Tango::UNKNOWN;
				}
				if (myacq->run_flag == false)
					break;							
			}
			/*********** end disable bpm fel_02 ***********/

			/************* enable bpm fel_01 *************/
			for (int i = 0; i < myacq->num_bpm_fel01; i++) {
				myacq->switch_fel01_counter++;
				try {	
					myacq->bpm_fel01[i].bpmdevice->command_inout("Enable");
					omni_thread::sleep(1,500000000);
					myacq->bpm_fel01[i].bpmdevice->read_attribute("State") >> myacq->bpm_fel01[i].state;
				}
				catch(Tango::DevFailed e) {
					myacq->bpm_fel01[i].state = Tango::UNKNOWN;										
				}	
				if (myacq->run_flag == false)
					break;
			}myacq->switch_fel01 = true;
			/*********** end enable bpm fel_01 ***********/

		}
		local_force_fel01_flag = 0;
		/************* END ENABLE FEL01 *************/	


		/************* ENABLE FEL02 *************/
		local_force_fel02_flag = myacq->force_fel02_flag;
		if (myacq->force_fel02_flag) myacq->force_fel02_flag = 0;		
		if (cmd_fel02_enable || local_force_fel02_flag) {
			myacq->switch_running = true;
			myacq->switch_fel01_counter = 0;
			myacq->switch_fel02_counter = 0;
			myacq->switch_fel01 = false;
			bool enable = true;
			/************* disable cbpm fel_01 *************/
			for (int i = 0; i < myacq->num_cbpm_fel01; i++) {
				try {
					/* recover from lock */
					if(myacq->cbpm_fel01[i].state == Tango::ALARM) {
						myacq->cbpm_fel01[i].cbpmdevice->command_inout("Init");
						omni_thread::sleep(1,500000000);
					}
					Tango::DeviceAttribute fel02enabled("Fel02Enabled", enable);
					myacq->cbpm_fel01[i].cbpmdevice->write_attribute(fel02enabled);
					omni_thread::sleep(0,150000000);
					myacq->cbpm_fel01[i].cbpmdevice->read_attribute("State") >> myacq->cbpm_fel01[i].state;
				}
				catch(Tango::DevFailed e) {
					myacq->cbpm_fel01[i].state = Tango::UNKNOWN;										
				}	
				if (myacq->run_flag == false)
					break;
			}	
			omni_thread::sleep(2,0);
			for (int i = 0; i < myacq->num_cbpm_fel01; i++) {
				try {
					myacq->cbpm_fel01[i].cbpmdevice->read_attribute("State") >> myacq->cbpm_fel01[i].state;
				}
				catch(Tango::DevFailed e) {
					myacq->cbpm_fel01[i].state = Tango::UNKNOWN;										
				}			
				if (myacq->run_flag == false)
					break;			
			}	
			/*********** end disable cbpm fel_01 ***********/

			/************* enable cbpm fel_02 *************/
			for (int i = 0; i < myacq->num_cbpm_fel02; i++) {
				myacq->switch_fel02_counter++;
				try {
					myacq->cbpm_fel02[i].cbpmdevice->read_attribute("State") >> myacq->cbpm_fel02[i].state;
				}
				catch(Tango::DevFailed e) {}
				try {	
					myacq->cbpm_fel02[i].cbpmdevice->command_inout("Init");
					omni_thread::sleep(1,500000000);
					Tango::DeviceAttribute fel02enabled("Fel02Enabled", true);
					myacq->cbpm_fel02[i].cbpmdevice->write_attribute(fel02enabled);
					omni_thread::sleep(1,500000000);
					myacq->cbpm_fel02[i].cbpmdevice->read_attribute("State") >> myacq->cbpm_fel02[i].state;
					/*printf("Init %s\n",myacq->cbpm_fel02[i].devname.c_str());*/
				}
				catch(Tango::DevFailed e) {
					myacq->cbpm_fel02[i].state = Tango::UNKNOWN;										
				}	
				if (myacq->run_flag == false)
					break;
			}
			/*********** end enable cbpm fel_02 ***********/

			/************* disable bpm fel_01 *************/
			for (int i = 0; i < myacq->num_bpm_fel01; i++) {
				try {
					myacq->bpm_fel01[i].bpmdevice->command_inout("Disable");
					omni_thread::sleep(1,500000000);
					myacq->bpm_fel01[i].bpmdevice->read_attribute("State") >> myacq->bpm_fel01[i].state;
				}
				catch(Tango::DevFailed e) {
					myacq->bpm_fel01[i].state = Tango::UNKNOWN;
				}
				if (myacq->run_flag == false)
					break;							
			}
			/*********** end disable bpm fel_01 ***********/

			/************* enable bpm fel_02 *************/
			for (int i = 0; i < myacq->num_bpm_fel02; i++) {
				myacq->switch_fel02_counter++;
				try {	
					myacq->bpm_fel02[i].bpmdevice->command_inout("Enable");
					omni_thread::sleep(1,500000000);
					myacq->bpm_fel02[i].bpmdevice->read_attribute("State") >> myacq->bpm_fel02[i].state;
				}
				catch(Tango::DevFailed e) {
					myacq->bpm_fel02[i].state = Tango::UNKNOWN;										
				}	
				if (myacq->run_flag == false)
					break;
			}
			/*********** end enable bpm fel_02 ***********/

		}				
		local_force_fel02_flag = 0;
		/************* END ENABLE FEL02 *************/	



		gettimeofday(&loop_end, NULL);	
		
		/* time elapsed from the start of the cycle */
		loop_time_len = (loop_end.tv_sec -loop_start.tv_sec) * 1000000 +
			(loop_end.tv_usec -loop_start.tv_usec);
		/* time to sleep */	
		loop_time_diff = myacq->checkPeriod*1000000 - loop_time_len;
		for (int i = 0; i < (int) (loop_time_diff/100000); i++) {
			usleep(100000);
			if (myacq->run_flag == false)
				break;
		}

	} /* while */

	myacq->run_thread = false;

}



}
